from collections.abc import Callable
from typing import cast

import yaml

from copicat.args import *
from copicat.utils import *

_all_ = ["main"]


def main():
    # print(args)
    # exit()
    match args.sub:
        case "in":
            iter_config(copy_in)
        case "out":
            iter_config(copy_out)
        case "type":
            type_sub()


def type_sub():
    assert args.sub == "type"
    if args.source.is_file():
        dst = args.dest / args.source.name
        copy_or_link_file(
            args.source,
            dst,
            args.hard_link,
            args.mimes,
            args.extensions,
            dry_types=args.dry_types,
        )
    elif args.source.is_dir() and args.dest.is_dir():
        if args.keep_structure:
            copy_dir_structure(
                args.source,
                args.dest,
                args.hard_link,
                args.mimes,
                args.extensions,
                args.dry_types,
            )
        else:
            copy_top_dir(
                args.source,
                args.dest,
                args.hard_link,
                args.mimes,
                args.extensions,
                args.dry_types,
            )


def iter_config(copy_routine: Callable[[Item], None]):
    assert args.sub == "in" or args.sub == "out"
    with open(args.config) as conf:
        config = cast(Schema, yaml.safe_load(conf))
        for item in loop_files(config):
            copy_routine(item)


def copy_in(item: Item):
    assert args.sub == "in"
    try:
        print(f'copy: {args.mode} {args.owner}:{args.group} "{item.dst}"->"{item.src}"')
        if not args.dry_run:
            uid = ensure_uid(args.owner)
            gid = ensure_gid(args.group)
            mode = convert_mode(args.mode)
            if item.dst.is_dir():
                copytree(item.dst, item.src, uid, gid, mode)
            else:
                copy_and_own(item.dst, item.src, uid, gid, mode)
    except Exception as ex:
        print(ex)


def copy_out(item: Item):
    try:
        print(
            f'copy: {item.octal} {item.owner}:{item.group} "{item.src}"->"{item.dst}"'
        )
        if not args.dry_run:
            if item.src.is_dir():
                copytree(item.src, item.dst, item.uid, item.gid, item.mode)
            else:
                copy_and_own(item.src, item.dst, item.uid, item.gid, item.mode)
    except Exception as ex:
        print(ex)


if __name__ == "__main__":
    main()



from itertools import chain
import os
import shutil
from collections.abc import Iterable
from grp import getgrnam
from pathlib import Path
from pwd import getpwnam
from typing import NamedTuple
import mimetypes

from copicat.args import args

__all__ = [
    "ensure_uid",
    "ensure_gid",
    "Schema",
    "Item",
    "loop_files",
    "copy_and_own",
    "copytree",
    "copy_dir_structure",
    "copy_top_dir",
    "convert_mode",
    "copy_or_link_file",
    "better_guess",
]

type Schema = dict[str | int, dict[str | int, dict[int, dict[str, str]]]]


class Item(NamedTuple):
    owner: str | int
    group: str | int
    octal: int
    uid: int
    gid: int
    mode: int
    src: Path
    dst: Path


def ensure_uid(owner: str | int) -> int:
    if isinstance(owner, int):
        return owner
    else:
        return getpwnam(owner).pw_uid


def ensure_gid(group: str | int) -> int:
    if isinstance(group, int):
        return group
    else:
        return getgrnam(group).gr_gid


def convert_mode(mode: int) -> int:
    return int(str(mode), base=8)


def loop_files(config: Schema) -> Iterable[Item]:
    for owner, next in config.items():
        try:
            uid = ensure_uid(owner)
            for group, next in next.items():
                try:
                    gid = ensure_gid(group)
                    for octal, next in next.items():
                        for src, dst in next.items():
                            mode = convert_mode(octal)
                            yield Item(
                                owner,
                                group,
                                octal,
                                uid,
                                gid,
                                mode,
                                Path(src),
                                Path(dst),
                            )
                except Exception as ex:
                    print(ex)
        except Exception as ex:
            print(ex)


def copy_and_own(src: Path, dst: Path, uid: int, gid: int, mode: int):
    shutil.copy(src, dst)
    os.chown(dst, uid, gid)
    os.chmod(dst, mode & 0o7777)


def copytree(src: Path, dst: Path, uid: int, gid: int, mode: int):
    shutil.copytree(src, dst, dirs_exist_ok=True)
    for dir, dirs, files in dst.walk():
        for name in dirs:
            dest = dir / name
            os.chown(dest, uid, gid)
            os.chmod(
                dest,
                (mode >> 12 if mode >= 1 << 12 else ((mode & 0o7777) | 0o111)),
            )

        for name in files:
            dest = dir / name
            os.chown(dest, uid, gid)
            os.chmod(dest, mode & 0o7777)


def copy_top_dir(
    src: Path,
    dst: Path,
    hard_link: bool,
    mime: list[str],
    extension: list[str],
    dry_types: bool,
):
    assert src.is_dir()
    for item in src.iterdir():
        copy_or_link_file(
            item, dst / item.name, hard_link, mime, extension, dry_types=dry_types
        )


def copy_dir_structure(
    src: Path,
    dst: Path,
    hard_link: bool,
    mimes: list[str],
    extensions: list[str],
    dry_types: bool,
):
    assert src.is_dir()
    base = dst / src.name
    for dir, dirs, files in src.walk():
        # print(f"{dir}, {files}, {dirs}")
        inner = dir.relative_to(src)
        for file in files:
            copy_or_link_file(
                dir / file,
                base / inner / file,
                hard_link,
                mimes,
                extensions,
                dry_types=dry_types,
                parents=True,
            )


def copy_or_link_file(
    src: Path,
    dst: Path,
    hard_link: bool,
    mimes: list[str],
    extensions: list[str],
    dry_types: bool = False,
    parents: bool = False,
):
    try:
        actual = better_guess(src)
        if dry_types:
            print(f'"{src}": {actual}')
        elif match_file(actual, src.suffix, mimes, extensions):
            if parents:
                os.makedirs(dst.parent, exist_ok=True)
            if hard_link:
                print(f'link: "{src}"->"{dst}"')
                if not args.dry_run:
                    os.link(src, dst)
            else:
                print(f'copy: "{src}"->"{dst}"')
                if not args.dry_run:
                    shutil.copy(src, dst)
    except Exception as ex:
        print(ex)


def match_file(
    actual_mime: str, actual_extension: str, mimes: list[str], extensions: list[str]
):
    return any(
        chain(
            (
                ("/" in mime and actual_mime == mime)
                or actual_mime.startswith(mime + "/")
                for mime in mimes
            ),
            (actual_extension == extension for extension in extensions),
        )
    )


def better_guess(path: os.PathLike) -> str:
    return mimetypes.guess_file_type(path, strict=False)[0] or "text/plain"



import argparse
import os
from pathlib import Path
from typing import Literal, NamedTuple, cast

__all__ = ["args", "Args", "InArgs", "OutArgs", "TypeArgs"]


class InArgs(NamedTuple):
    sub: Literal["in"]
    config: Path
    owner: str | int
    group: str | int
    mode: int
    dry_run: bool


class OutArgs(NamedTuple):
    sub: Literal["out"]
    config: Path
    dry_run: bool


class TypeArgs(NamedTuple):
    sub: Literal["type"]
    source: Path
    dest: Path
    keep_structure: bool
    hard_link: bool
    mimes: list[str]
    extensions: list[str]
    dry_types: bool
    dry_run: bool


type Args = InArgs | OutArgs | TypeArgs

_parser = argparse.ArgumentParser("copicat")
_subparsers = _parser.add_subparsers(
    title="subcommands", dest="sub", required=True, help="copicat subcommands"
)

_parser_out = _subparsers.add_parser(
    "out", help="Copy files from source to target location"
)
_parser_in = _subparsers.add_parser(
    "in", help="Copy files from target to source location"
)
_parser_type = _subparsers.add_parser(
    "type",
    help="Copy files from source to destination location based on mime type or extension",
)

_parser_out.add_argument(
    "config",
    nargs="?",
    help='Path to YAML config file. Default is "copicat.yml"',
    default="copicat.yml",
    type=Path,
)

_parser_in.add_argument(
    "config",
    nargs="?",
    help='Path to YAML config file. Default is "copicat.yml"',
    default="copicat.yml",
    type=Path,
)

_parser_in.add_argument(
    "owner",
    nargs="?",
    help="Username or UID of the owner of the files",
    default=os.getuid(),
)

_parser_in.add_argument(
    "group",
    nargs="?",
    help="Group name or GID of the group of the files",
    default=os.getgid(),
)

_parser_in.add_argument(
    "mode",
    nargs="?",
    type=int,
    help="Mode of the crated files in the same format as the config file",
    default="660",
)

_parser_type.add_argument("source", help="Source directory or file", type=Path)

_parser_type.add_argument("dest", help="Destination directory", type=Path)

_parser_type.add_argument(
    "-m",
    "--mimes",
    metavar="MIME",
    nargs="*",
    help='Half or full of mime types to copy (i.e.: both "video" and "video/x-matroska" will match a file with mimetype "video/x-matroska")',
    default=[],
)

_parser_type.add_argument(
    "-e",
    "--extensions",
    metavar="EXTENSION",
    nargs="*",
    help='Extensions with point of the files to copy (i.e.: ".srt" will match "subs.srt")',
    default=[],
)

_parser_type.add_argument(
    "-k",
    "--keep-structure",
    help="Keep the directory structure in the destination. Without this, only the files from the first level will be copied",
    action="store_true",
)

_parser_type.add_argument(
    "-H",
    "--hard-link",
    help="Hard-link files into destination instead of copying",
    action="store_true",
)

_parser_type.add_argument(
    "-T",
    "--dry-types",
    help="Just print the types of all candidate source files. Different implementations might identify types differently",
    action="store_true",
)


#### dry run on all parsers

_parser_type.add_argument(
    "-D",
    "--dry-run",
    help="Print the operation without performing any action",
    action="store_true",
)

_parser_in.add_argument(
    "-D",
    "--dry-run",
    help="Print the operation without performing any action",
    action="store_true",
)

_parser_out.add_argument(
    "-D",
    "--dry-run",
    help="Print the operation without performing any action",
    action="store_true",
)

####

args = cast(Args, _parser.parse_args())
